from django.contrib.sites.shortcuts import get_current_site
from django.core.mail import send_mail
from django.template.loader import get_template, render_to_string
from django.conf import settings
from . import settings as stagedoor_settings
from .models import AuthToken, Email, PhoneNumber, generate_token_string


def email_login_link(request, token):
    current_site = get_current_site(request)

    # Send the link by email.
    send_mail(
        subject=f"Here's your login to {stagedoor_settings.SITE_NAME}",
        message=render_to_string(
            stagedoor_settings.EMAIL_TXT_TEMPLATE,
            {
                "current_site": current_site,
                "token": token.token,
                "site_name": stagedoor_settings.SITE_NAME,
                "support_email": stagedoor_settings.SUPPORT_EMAIL,
            },
            request=request,
        ),
        from_email=stagedoor_settings.DEFAULT_FROM_EMAIL,
        recipient_list=[token.email.email],
        html_message=get_template(stagedoor_settings.EMAIL_HTML_TEMPLATE).render(
            {
                "current_site": current_site,
                "token": token.token,
                "site_name": stagedoor_settings.SITE_NAME,
                "support_email": stagedoor_settings.SUPPORT_EMAIL,
            },
        ),
        fail_silently=False,
    )


def sms_login_link(request, token):
    current_site = get_current_site(request)
    if (
        hasattr(settings, "TWILIO_ACCOUNT_SID")
        and hasattr(settings, "TWILIO_AUTH_TOKEN")
        and hasattr(settings, "TWILIO_NUMBER")
        and settings.TWILIO_ACCOUNT_SID is not None
        and settings.TWILIO_AUTH_TOKEN is not None
        and settings.TWILIO_NUMBER is not None
    ):
        from twilio.rest import Client

        client = Client(settings.TWILIO_ACCOUNT_SID, settings.TWILIO_AUTH_TOKEN)
        message = client.messages.create(
            body=f"Your {stagedoor_settings.SITE_NAME} code is {token.token}\n\nGo to https://{current_site.domain}/auth/token to login.",
            from_=settings.TWILIO_NUMBER,
            to=str(token.phone_number.phone_number),
        )



from django.contrib.auth import get_user_model

from . import settings as stagedoor_settings
from .models import AuthToken
from .models import generate_token_string


class StageDoorBackendMixin(object):
    def get_user(self, user_id):
        """Get a user by their primary key."""
        User = get_user_model()
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None

    def authenticate(self, request, token=None):
        """Authenticate a user given a token"""
        AuthToken.delete_stale()

        token = AuthToken.objects.filter(token=token).first()
        if not token:
            return

        if stagedoor_settings.SINGLE_USE_LINK:
            token.delete()

        User = get_user_model()

        user_args = {}

        if "username" in [
            field.name for field in User._meta.get_fields(include_hidden=True)
        ]:
            user_args["username"] = f"u{generate_token_string()[:8]}"

        user, created = User.objects.get_or_create(**user_args)

        if token.next_url:
            user._stagedoor_next_url = token.next_url

        return user


class EmailTokenBackend(StageDoorBackendMixin):
    def authenticate(self, request, token=None):
        token = AuthToken.objects.filter(token=token, email__isnull=False).first()
        if not token:
            return

        email = token.email
        if (
            email.potential_user
            and request.user.is_authenticated
            and request.user == email.potential_user
        ):
            email.user = email.potential_user
            email.potential_user = None
            email.save()

        if email.user:
            user = email.user
        else:
            user = super(EmailTokenBackend, self).authenticate(request, token.token)

        if user:
            if token.next_url:
                user._stagedoor_next_url = token.next_url
            User = get_user_model()

            if "email" in [
                field.name for field in User._meta.get_fields(include_hidden=True)
            ]:
                user.email = email.email
            if email:
                email.user = user
                email.save()
            if stagedoor_settings.SINGLE_USE_LINK:
                token.delete()
            user.save()
        return user


class SMSTokenBackend(StageDoorBackendMixin):
    def authenticate(self, request, token=None):
        """Authenticate a user given a signed token from email."""
        token = AuthToken.objects.filter(
            token=token, phone_number__isnull=False
        ).first()
        if not token:
            return

        phone_number = token.phone_number

        if (
            phone_number.potential_user
            and request.user.is_authenticated
            and phone_number.potential_user == request.user
        ):
            phone_number.user = phone_number.potential_user
            phone_number.potential_user = None
            phone_number.save()

        if phone_number.user:
            user = phone_number.user
        else:
            user = super(SMSTokenBackend, self).authenticate(request, token.token)

        if user:
            if token.next_url:
                user._stagedoor_next_url = token.next_url
            User = get_user_model()

            if "phone_number" in [
                field.name for field in User._meta.get_fields(include_hidden=True)
            ]:
                user.phone_number = phone_number.phone_number
                user.save()

            if phone_number:
                phone_number.user = user
                phone_number.save()
            if stagedoor_settings.SINGLE_USE_LINK:
                token.delete()
            user.save()
        return user



from django.urls import path

from . import views


app_name = "stagedoor"
urlpatterns = [
    path("login", views.login_post, name="login"),
    path("login/<str:token>", views.token_login, name="token-login"),
    path("logout", views.logout, name="logout"),
    path("token", views.token_post, name="token-post"),
]



import logging
import random
from datetime import timedelta

from django.contrib.auth import get_user_model
from django.db import models
from django.utils.timezone import now

from django.conf import settings

from phonenumber_field.modelfields import PhoneNumberField

from . import settings as stagedoor_settings

logger = logging.getLogger(__name__)


def generate_token_string(email=False, sms=False):
    if email and sms:
        logger.error("Tried to generate a token for email and sms")
        return
    token_length = stagedoor_settings.EMAIL_TOKEN_LENGTH
    charset = "abcdefghijkmnopqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ123456789"

    if sms:
        charset = "123456789"
        token_length = stagedoor_settings.SMS_TOKEN_LENGTH

    return "".join([random.choice(charset) for _ in range(token_length)])


def generate_sms_token(phone_number, next_url, user=None):
    token_string = generate_token_string(sms=True)
    phone_number, created = PhoneNumber.objects.get_or_create(
        phone_number=phone_number,
    )
    if (not user or not user.is_authenticated) or created:
        token = AuthToken.objects.create(
            token=token_string, phone_number=phone_number, next_url=next_url,
        )
        return token
    else:
        if phone_number.user and phone_number.user != user:
            return None
        else:
            phone_number.potential_user = user
            phone_number.save()
            token = AuthToken.objects.create(
                token=token_string, phone_number=phone_number, next_url=redirect_url,
            )
            return token


def generate_email_token(email, next_url, user=None):
    token_string = generate_token_string(email=True)
    email, created = Email.objects.get_or_create(email=email)
    if (not user or not user.is_authenticated) or created:
        token = AuthToken.objects.create(
            token=token_string, email=email, next_url=next_url
        )
        return token

    if email.user and email.user != user:
        return None
    else:
        email.potential_user = user
        email.save()
        token = AuthToken.objects.create(
            token=token_string, email=email, next_url=redirect_url
        )
        return token


class Email(models.Model):

    email = models.EmailField(unique=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="stagedoor_user_email",
    )
    potential_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="stagedoor_potential_user_email",
    )

    def __str__(self):
        return f"{self.email}: {self.user} (Maybe: {self.potential_user})"


class PhoneNumber(models.Model):

    phone_number = PhoneNumberField(
        help_text="Must include international prefix - e.g. +1 555 555 55555",
        unique=True,
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="stagedoor_user_phone_number",
    )
    potential_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="stagedoor_potential_user_phone_number",
    )

    def __str__(self):
        return f"{self.phone_number}: {self.user} (Maybe: {self.potential_user})"


class AuthToken(models.Model):
    token = models.CharField(max_length=200, db_index=True)
    timestamp = models.DateTimeField(auto_now_add=True, db_index=True)
    email = models.ForeignKey(Email, blank=True, null=True, on_delete=models.CASCADE)
    phone_number = models.ForeignKey(
        PhoneNumber, blank=True, null=True, on_delete=models.CASCADE
    )
    next_url = models.CharField(max_length=2000, blank=True)

    @classmethod
    def delete_stale(cls):
        """Delete stale tokens, ie tokens that are more than TOKEN_DURATION seconds older."""
        cls.objects.filter(
            timestamp__lt=now() - timedelta(seconds=stagedoor_settings.TOKEN_DURATION)
        ).delete()

    def __str__(self):
        return self.timestamp.strftime("%Y-%m-%d %H:%M:%S")



from django import forms
from django.contrib import messages
from django.contrib.auth import authenticate
from django.contrib.auth import login as django_login
from django.contrib.auth import logout as django_logout
from django.contrib.auth.decorators import login_required
from django.shortcuts import redirect, render
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views.decorators.http import require_http_methods

from phonenumber_field.formfields import PhoneNumberField

from django.core.validators import validate_email
from django.core.validators import ValidationError

from phonenumber_field.validators import validate_international_phonenumber

from . import settings as stagedoor_settings
from .helpers import email_login_link, sms_login_link
from .models import (
    AuthToken,
    Email,
    PhoneNumber,
    generate_email_token,
    generate_sms_token,
    generate_token_string,
)


class LoginForm(forms.Form):
    """The form for the login page."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        label = "Your contact information"
        if stagedoor_settings.ENABLE_SMS:
            label = "Your phone number, ie +15555555555"
            self.fields["phone_number"] = PhoneNumberField(
                label="Your phone number", required=False
            )
        if stagedoor_settings.ENABLE_EMAIL:
            label = "Your email"
            self.fields["email"] = forms.EmailField(
                label="Your email address", required=False
            )
        if stagedoor_settings.ENABLE_SMS and stagedoor_settings.ENABLE_EMAIL:
            label = "Your phone number or email"

        self.fields["contact"] = contact = forms.CharField(label=label, required=False)

    def clean(self):
        cleaned_data = super().clean()
        form_email = cleaned_data.get("email")
        form_phone_number = cleaned_data.get("phone_number")

        try:
            contact_email = cleaned_data.get("contact")
            validate_email(contact_email)
        except ValidationError:
            contact_email = None
        try:
            contact_phone = cleaned_data.get("contact")
            validate_international_phonenumber(contact_phone)
        except ValidationError:
            contact_phone = None

        email = contact_email if not form_email else form_email
        phone_number = contact_phone if not form_phone_number else form_phone_number

        if not email and not phone_number:
            raise forms.ValidationError(
                "Please use a valid email address or phone number."
            )

        self.cleaned_data["phone_number"] = phone_number
        self.cleaned_data["email"] = email
        return self.cleaned_data


@require_http_methods(["POST"])
def login_post(request):
    """Process the submission of the form with the user's email and mail them a link."""
    form = LoginForm(request.POST)
    if not form.is_valid():
        messages.error(
            request, _("Please use a valid email address or phone number."),
        )
        return redirect(stagedoor_settings.LOGIN_URL)

    email = form.cleaned_data["email"]
    phone_number = form.cleaned_data["phone_number"]

    redirect_url = ""
    if request.GET and request.GET.get("next"):
        redirect_url = request.GET.get("next", "")

    if email:
        token = generate_email_token(email, next_url=redirect_url, user=request.user)
        if token:
            email_login_link(request=request, token=token)
            messages.success(
                request, _("Check your email to log in!"),
            )
            return redirect(reverse("stagedoor:token-post"))
        else:
            messages.error(request, _("A user with that email already exists."))
            return redirect(stagedoor_settings.LOGIN_URL)

    elif phone_number:
        token = generate_sms_token(
            phone_number, next_url=redirect_url, user=request.user
        )
        if token:
            sms_login_link(request=request, token=token)
            messages.success(
                request, _("Check your text messages to log in!"),
            )
            return redirect(reverse("stagedoor:token-post"))
        else:
            messages.error(request, _("A user with that phone number already exists."))
            return redirect(stagedoor_settings.LOGIN_URL)

    return redirect(stagedoor_settings.LOGIN_URL)


def process_token(request, token):
    user = authenticate(request, token=token)
    if user is None:
        messages.error(
            request,
            _(
                "The login link is invalid or has expired, or you are not allowed to "
                "log in. Please try again."
            ),
        )
        return redirect(stagedoor_settings.LOGIN_URL)

    if hasattr(user, "_stagedoor_next_url"):
        # Get the next URL from the user object, if it was set by our custom `authenticate`.
        next_url = user._stagedoor_next_url

        # Remove the next URL from the user object.
        del user._stagedoor_next_url
    else:
        next_url = stagedoor_settings.LOGIN_REDIRECT

    if not request.user.is_authenticated:
        django_login(request, user)
    messages.success(request, _("Login successful."))
    return redirect(next_url)


def token_post(request):
    if request.POST:
        token = request.POST.get("token")
        return process_token(request, token)
    return render(request, template_name="stagedoor_token_input.html")


@require_http_methods(["GET"])
def token_login(request, token):
    """Validate the token the user submitted."""
    return process_token(request, token)


@login_required
def logout(request):
    django_logout(request)
    messages.success(request, _("You have been logged out."))
    return redirect(stagedoor_settings.LOGOUT_REDIRECT)



from django.contrib import admin

from .models import AuthToken, Email, PhoneNumber


@admin.register(Email)
class EmailAdmin(admin.ModelAdmin):
    list_display = ("email", "user", "potential_user")


@admin.register(PhoneNumber)
class PhoneNumberAdmin(admin.ModelAdmin):
    list_display = ("phone_number", "user", "potential_user")


@admin.register(AuthToken)
class AuthTokenAdmin(admin.ModelAdmin):
    list_display = ["email", "phone_number", "timestamp", "next_url"]
    ordering = ["-timestamp"]



from django.conf import settings


TOKEN_DURATION = getattr(settings, "STAGEDOOR_TOKEN_DURATION", 30 * 60)

EMAIL_TOKEN_LENGTH = getattr(settings, "STAGEDOOR_EMAIL_TOKEN_LENGTH", 8)

SMS_TOKEN_LENGTH = getattr(settings, "STAGEDOOR_SMS_TOKEN_LENGTH", 6)

LOGIN_URL = getattr(settings, "STAGEDOOR_LOGIN_URL", settings.LOGIN_URL)

LOGIN_REDIRECT = getattr(
    settings, "STAGEDOOR_LOGIN_REDIRECT", settings.LOGIN_REDIRECT_URL
)
LOGOUT_REDIRECT = getattr(
    settings,
    "STAGEDOOR_LOGOUT_REDIRECT",
    settings.LOGOUT_REDIRECT_URL or settings.LOGIN_REDIRECT_URL,
)

DEFAULT_FROM_EMAIL = getattr(
    settings, "STAGEDOOR_DEFAULT_FROM_EMAIL", settings.DEFAULT_FROM_EMAIL
)

SUPPORT_EMAIL = getattr(
    settings, "STAGEDOOR_SUPPORT_EMAIL", settings.DEFAULT_FROM_EMAIL
)

SINGLE_USE_LINK = getattr(settings, "STAGEDOOR_SINGLE_USE_LINK", False)

EMAIL_HTML_TEMPLATE = getattr(
    settings, "STAGEDOOR_EMAIL_HTML_TEMPLATE", "stagedoor_email.html"
)
EMAIL_TXT_TEMPLATE = getattr(
    settings, "STAGEDOOR_EMAIL_TXT_TEMPLATE", "stagedoor_email.txt"
)

ALLOW_MULTIPLE_EMAILS = getattr(settings, "STAGEDOOR_ALLOW_MULTIPLE_EMAILS", False)

ALLOW_MULTIPLE_PHONE_NUMBERS = getattr(
    settings, "STAGEDOOR_ALLOW_MULTIPLE_PHONE_NUMBERS", False
)

ENABLE_SMS = getattr(settings, "STAGEDOOR_ENABLE_SMS_OVERRIDE", False) or (
    hasattr(settings, "TWILIO_ACCOUNT_SID")
    and hasattr(settings, "TWILIO_AUTH_TOKEN")
    and hasattr(settings, "TWILIO_NUMBER")
)

ENABLE_EMAIL = getattr(settings, "STAGEDOOR_ENABLE_EMAIL_OVERRIDE", False) or (
    hasattr(settings, "EMAIL_HOST")
    and hasattr(settings, "EMAIL_HOST_USER")
    and hasattr(settings, "EMAIL_HOST_PASSWORD")
)

SITE_NAME = getattr(settings, "STAGEDOOR_SITE_NAME", "Django")



from rest_framework import serializers

from .models import PhoneNumber
from .models import Email


class EmailSerializer(serializers.ModelSerializer):
    class Meta:
        model = Email
        fields = ("email",)


class PhoneNumberSerializer(serializers.ModelSerializer):
    class Meta:
        model = PhoneNumber
        fields = ("phone_number",)



import os
from setuptools import setup

with open("README.md", "r") as fh:
    README = fh.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="django-stagedoor",
    version="0.0.1",
    zip_safe=False,
    packages=["stagedoor"],
    author="Philip James",
    author_email="phildini@phildini.net",
    license="Apache License",
    description="A Django app for passwordless login with SMS and Email",
    include_package_data=True,
    long_description=README,
    long_description_content_type="text/markdown",
    install_requires=[
        'twilio>=6.40.0,<6.50.0',
        'django>=2.2,<4.0',
        'django-phonenumber-field>=4.0.0,<4.1.0',
        'phonenumbers>=8.12.0,<8.13',
    ],
    url="https://github.com/galaxybrainco/django-stagedoor",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet :: WWW/HTTP",
    ],
    python_requires='>=3.6',
)